@@ -13,6 +13,7 @@ import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.CopyOnWriteArrayList;
+import java.util.concurrent.ScheduledExecutorService;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.locks.Lock;
@@ -45,6 +46,7 @@ public class ResponseTracker implements Runnable {
     private ConcurrentMap<String, List<JsonNode>> hostToId = new ConcurrentHashMap<>();
     private Queue<JsonNode> queue = new ConcurrentLinkedQueue<>();
     private final Lock lock = new ReentrantLock();
+    private ScheduledExecutorService executorService;
 
     public ResponseTracker() {
         this.isTracking = new AtomicBoolean(true);
@@ -83,7 +85,7 @@ public class ResponseTracker implements Runnable {
             this.queue.add(id);
             nodes.add(id);
             nodes = this.hostToId.putIfAbsent(tracking.getClient().getClientId(), nodes);
-            if (nodes != null && !nodes.contains(id)) {
+            if (nodes != null) {
                 nodes.add(id);
             }
         }
@@ -94,37 +96,33 @@ public class ResponseTracker implements Runnable {
         try {
             while (this.isTracking.get()) {
                 TimeUnit.MILLISECONDS.sleep(TRACKING_TIMEOUT);
-                loop();
+                for (JsonNode id : queue) {
+                    if (!this.runningCalls.containsKey(id)) {
+                        removeRequestFromTracking(id);
+                        continue;
+                    }
+                    ResponseTracking tracking = this.map.get(id);
+                    if (System.currentTimeMillis() >= tracking.getTimeout()) {
+                        RetryContext context = tracking.getContext();
+                        context.decreaseAttempts();
+                        if (context.getNumberOfAttempts() <= 0) {
+                            handleFailure(tracking, id);
+                            continue;
+                        }
+                        try {
+                            tracking.getClient().sendMessage(jsonToByteArray(tracking.getRequest().toJson()));
+                        } catch (ClientConnectionException e) {
+                            handleFailure(tracking, id);
+                        }
+                        tracking.setTimeout(getTimeout(context.getTimeout(), context.getTimeUnit()));
+                    }
+                }
             }
         } catch (InterruptedException e) {
             log.warn("Tracker thread intrreupted");
         }
     }
 
-    protected void loop() {
-        for (JsonNode id : queue) {
-            if (!this.runningCalls.containsKey(id)) {
-                removeRequestFromTracking(id);
-                continue;
-            }
-            ResponseTracking tracking = this.map.get(id);
-            if (System.currentTimeMillis() >= tracking.getTimeout()) {
-                RetryContext context = tracking.getContext();
-                context.decreaseAttempts();
-                if (context.getNumberOfAttempts() <= 0) {
-                    handleFailure(tracking, id);
-                    continue;
-                }
-                try {
-                    tracking.getClient().sendMessage(jsonToByteArray(tracking.getRequest().toJson()));
-                } catch (ClientConnectionException e) {
-                    handleFailure(tracking, id);
-                }
-                tracking.setTimeout(getTimeout(context.getTimeout(), context.getTimeUnit()));
-            }
-        }
-    }
-
     public void close() {
         this.isTracking.set(false);
     }
@@ -136,15 +134,33 @@ public class ResponseTracker implements Runnable {
         }
     }
 
+    public void setExecutorService(ScheduledExecutorService executorService) {
+        this.executorService = executorService;
+    }
+
     private void remove(ResponseTracking tracking, JsonNode id, JsonRpcResponse response) {
         try (LockWrapper wrapper = new LockWrapper(this.lock)) {
             JsonRpcCall call = this.runningCalls.remove(id);
+            boolean callbackNotified = false;
             if (call != null) {
                 call.addResponse(response);
+                if (call.getCallback() != null && executorService != null) {
+                    callbackNotified = true;
+                    executorService.schedule(() -> call.getCallback().onFailure(mapValues(response.getError())),
+                            0,
+                            TimeUnit.SECONDS);
+                }
             }
             removeRequestFromTracking(id);
-            if (tracking != null && tracking.getClient() != null) {
+            if (!callbackNotified && tracking != null && tracking.getClient() != null) {
                 tracking.getCall().addResponse(response);
+                if (tracking.getCall().getCallback() != null && executorService != null) {
+                    executorService.schedule(() -> tracking.getCall()
+                            .getCallback()
+                            .onFailure(mapValues(response.getError())),
+                            0,
+                            TimeUnit.SECONDS);
+                }
             }
         }
     }
@@ -161,7 +177,8 @@ public class ResponseTracker implements Runnable {
                 removeNodes(this.hostToId.get(code), errorResponse);
             } else {
                 String hostname = code.substring(0, code.indexOf(":"));
-                this.hostToId.keySet().stream().filter(key -> key.startsWith(hostname))
+                this.hostToId.keySet().stream()
+                        .filter(key -> key.startsWith(hostname))
                         .forEach(key -> removeNodes(this.hostToId.get(key), errorResponse));
             }
         }
@@ -172,8 +189,4 @@ public class ResponseTracker implements Runnable {
                 .filter(id -> !NullNode.class.isInstance(id))
                 .forEach(id -> remove(this.map.get(id), id, errorResponse));
     }
-
-    protected Map<String, List<JsonNode>> getHostMap() {
-        return this.hostToId;
-    }
 }
